(* EMX.icl -- included by SysDep.icl *)


(* Lo-level driver routines for DOS and OS/2 systems with VGA display *)


(******************************************************************************)
(*                                                                            *)
(*  This file is part of BO5, the user interface library for GNU Pascal.      *)
(*                                                                            *)
(*  BO5 will be distributed as free software (probably under the terms of     *)
(*  the GNU Library General Public License) as soon as it will have reached   *)
(*  some state of completion.  Up to that time, you are allowed to use BO5    *)
(*  for non-commercial purposes only.  You may distribute verbatim copies of  *)
(*  BO5 as a whole, but only without -- even a nominal -- fee.  You may not   *)
(*  distribute parts or changed versions of BO5.  I stress that there is      *)
(*  absolutely NO WARRANTY for BO5 -- any risk of using BO5 is with you!      *)
(*                                                                            *)
(*  Copyright (C) 1995  Peter Gerwinski, Duesseldorfer Str. 35                *)
(*                      (with TeX: D\"usseldorfer Str. 35)                    *)
(*                      45145 Essen, Germany                                  *)
(*                      e-mail: pege@mail.theo-phys.uni-essen.de              *)
(*                                                                            *)
(******************************************************************************)


(* Dear experienced lo-level programmer!       *)
(*                                             *)
(* Please don't cry when you read this source. *)
(* It was written rather hastily and will be   *)
(* optimized later on.                         *)
(*                                             *)
(* I have problems with the GNU inline assem-  *)
(* bler.  What is the correct way to use it?   *)
(* Any help welcome! -- PG                     *)


(*$ifdef SYSTOOLS *)  (* This goes into the Interface section of Tools *)


Var
  seg0040: Word value $0040;  (* BIOS variables *)
  segA000: Word value $A000;  (* Video memory *)
  segB000: Word value $B000;
  segB800: Word value $B800;


Procedure EMXinvalidOS;  (* announce a fatal error *)


(*$endif *)  (* SYSTOOLS *)


(*$ifdef TOOLS *)  (* This goes into the Implementation section of Tools *)


(* EMX lo-level memory access *)

Function MemAccess ( First, Last, Flags: Word ): Word;
AsmName '_memaccess';

(* In fact it returns a pointer.  But we convert it to Word. *)


Procedure EMXinvalidOS;

begin (* EMXinvalidOS *)
  FatalError ( 'EMX drivers: cannot identify operating system' );
end (* EMXinvalidOS *);


Procedure InitTools;

Var
  os2mode: Boolean;
  emxenv: ShortWord;


(*$W-*)  (* Suppress "return value not assigned" warning *)

Function SegToSel ( Seg: ShortWord ) Sel: Word;

begin (* SegToSel *)
  (* The GNU inline assembler seems not as comfortable as the   *)
  (* Borland inline assembler, and I don't yet fully understand *)
  (* how to use it.  Perhaps somebody can tell me? -- PG        *)
  asm ( 'movw %0, %%bx' : : 'rm' ( Seg ) );
  asm ( 'movw $0x0002, %ax' );
  asm ( 'int $0x31' );
  asm ( 'movl %%eax, %0' : : 'rm' ( Sel ) );
end (* SegToSel *);

(*$W+*)


begin (* InitTools *)
  (* As long as GPC cannot access external variables directly, :-( *)
  (* we must do the following stupid assignment manually with asm. *)
  asm ( 'movb (__osmode), %al' );
  asm ( 'movb %%al, %0' : : 'rm' ( os2mode ) );
  if os2mode then
    WhichSystem:= systemOS2
  else
    begin
      asm ( 'movw (__emx_env), %ax' );
      asm ( 'movb %%ax, %0' : : 'rm' ( emxenv ) );
      if emxenv and $0180 <> 0 then
        begin
          WhichSystem:= systemDPMI;
          seg0040:= SegToSel ( $0040 );
          segA000:= SegToSel ( $A000 );
          segB000:= SegToSel ( $B000 );
          segB800:= SegToSel ( $B800 );
        end (* if *)
      else
        begin
          WhichSystem:= systemDOS;
          seg0040:= MemAccess ( $00000, $00FFF, 0 );  (* read only *)
          segA000:= MemAccess ( $A0000, $AFFFF, 1 );  (* read/write *)
          segB000:= MemAccess ( $B0000, $BFFFF, 1 );
          segB800:= MemAccess ( $B8000, $BFFFF, 1 );  (* overlap *)
        end (* else *);
    end (* else *);
end (* InitTools *);


Procedure FiniTools;

begin (* FiniTools *)
  (* empty *)
end (* FiniTools *);


(*$endif *)  (* TOOLS *)


(*$ifdef EVENTS *)


(* OS/2 API keyboard access *)


Type


KbdInfo = record
  cb, fsMask, chTurnAround, fsInterim, fsState: ShortWord;
end (* KbdInfo *);


KbdKeyInfo = record
  chChar, chScan, fbStatus, bNlsShift: Byte;
  fsState: ShortWord;
  time: Word;
end (* KbdKeyInfo *);


Function KbdPeek ( Var Info: KbdKeyInfo;
                   Handle: ShortWord ): ShortWord;
AsmName 'KbdPeek';

Function KbdCharIn ( Var Info: KbdKeyInfo;
                     fWait, Handle: ShortWord ): ShortWord;
AsmName 'KbdCharIn';

Function KbdOpen ( Var Handle: ShortWord ): ShortWord;
AsmName 'KbdOpen';

Function KbdClose ( Handle: ShortWord ): ShortWord;
AsmName 'KbdClose';

Function KbdGetFocus ( fWait, Handle: ShortWord ): ShortWord;
AsmName 'KbdGetFocus';

Function KbdFreeFocus ( Handle: ShortWord ): ShortWord;
AsmName 'KbdFreeFocus';

Function KbdSetStatus ( Var Info: KbdInfo;
                        Handle: ShortWord ): ShortWord;
AsmName 'KbdSetStatus';


Var
  KbdHandle: ShortWord value 0;


Function KeyPressed: Boolean;

Var
  Info: KbdKeyInfo;
  result: Boolean;

begin (* KeyPressed *)
  case WhichSystem of
    systemDOS, systemDPMI:
      begin
        asm ( 'pushw %ax' );
        asm ( 'movb $0x11, %ah' );
        asm ( 'int $0x16' );
        asm ( 'popw %ax' );
        asm ( 'setne %0' : : 'rm' ( result ) );
        KeyPressed:= result;
      end (* systemDOS, systemDPMI *);
    systemOS2:
      begin
        KbdPeek ( Info, KbdHandle );
        KeyPressed:= Info.fbStatus and $40 <> 0;
      end (* systemOS2 *);
    otherwise:
      EMXinvalidOS;
  end (* case *);
end (* KeyPressed *);


Function ReadKey: Char;

Var
  Info: KbdKeyInfo;
  result: Char;

begin (* ReadKey *)
  case WhichSystem of
    systemDOS, systemDPMI:
      begin
        asm ( 'pushl %eax' );
        asm ( 'movb $0x10, %ah' );
        asm ( 'int $0x16' );
        asm ( 'movb %%al, %0' : : 'rm' ( result ) );
        asm ( 'popl %eax' );
        ReadKey:= result;
      end (* systemDOS, systemDPMI *);
    systemOS2:
      begin
        KbdCharIn ( Info, 0, KbdHandle );
        ReadKey:= chr ( Info.chChar );
      end (* systemOS2 *);
    otherwise:
      EMXinvalidOS;
  end (* case *);
end (* ReadKey *);


Procedure InitEvents;

Var
  Info: KbdInfo;

begin (* InitEvents *)
  if WhichSystem = systemOS2 then
    begin
      with Info do
        begin
          cb:= 10;
          fsMask:= $04;
        end (* with *);
      if ( KbdOpen ( KbdHandle ) <> 0 )
       or ( KbdGetFocus ( 0, KbdHandle ) <> 0 )
       or ( KbdSetStatus ( Info, 0 ) <> 0 ) then
        FatalError ( 'InitEvents (EMX): cannot initialize keyboard' );
    end (* if *);
end (* InitEvents *);


Procedure FiniEvents;

begin (* FiniEvents *)
  if WhichSystem = systemOS2 then
    begin
      KbdFreeFocus ( KbdHandle );
      KbdClose ( KbdHandle );
    end (* if *);
end (* FiniEvents *);


(*$endif *)  (* EVENTS *)


(*$ifdef DISPLAYS *)


(* OS/2 API video access *)


Type


VioCursorInfo = record
  yStart, cEnd, cx, Attr: ShortWord;
end (* VioCursorInfo *);


Function VioWrtCellStr ( Var B: Void; w, y, x: Integer;
                         Handle: ShortWord ): ShortWord;
AsmName 'VioWrtCellStr';

Function VioGetBuf ( Var addr, len: Word;
                     Handle: ShortWord ): ShortWord;
AsmName 'VioGetBuf';

Function VioSetCurPos ( y, x: Integer;
                        Handle: ShortWord ): ShortWord;
AsmName 'VioSetCurPos';

Function VioGetCurType ( Var Info: VioCursorInfo;
                         Handle: ShortWord ): ShortWord;
AsmName 'VioGetCurType';

Function VioSetCurType ( Var Info: VioCursorInfo;
                         Handle: ShortWord ): ShortWord;
AsmName 'VioSetCurType';


Var
  SegVideo: Word value 0;
  StartupCursorInfo: VioCursorInfo;


Constructor DisplayObj.Init;

Var
  VioBufAddr, VioBufLen: Word;

begin (* DisplayObj.Init *)
  Size.x:= 80;
  Size.y:= 25;
  CharSize.x:= 1;
  CharSize.y:= 1;
  if WhichSystem = systemOS2 then
    begin
      VioGetCurType ( StartupCursorInfo, 0 );
      if VioGetBuf ( VioBufAddr, VioBufLen, 0 ) = 0 then
        SegVideo:= VioBufAddr;  (* not used *)
    end (* if *)
  else
    SegVideo:= segB800;
end (* DisplayObj.Init *);


Destructor DisplayObj.Fini;

begin (* DisplayObj.Fini *)
  (* empty *)
end (* DisplayObj.Fini *);


Procedure DisplayObj.Clear ( Attr: Byte );

(* Attr specifies a background color, so we select *)
(* a reasonable foreground color.                  *)
(* This only concerns subsequent write calls the   *)
(* application program should not use anyway.      *)
(* But if a fatal error happens, the message will  *)
(* be visible.                                     *)

Var
  ForeGround, BackGround: Byte;

begin (* DisplayObj.Clear *)
  BackGround:= Attr and $0F;
  ForeGround:= ( Attr and $F0 ) shr 4;
  if ( ForeGround = black ) and ( BackGround < lightgray ) then
    ForeGround:= lightgray;
  WriteChars ( 0, 0, Size.x, Size.y, ' ',
               ForeGround or BackGround shl 4 );
end (* DisplayObj.Clear *);


Procedure DisplayObj.SetCursor ( x, y: Integer );

Var
  dx: ShortWord;

begin (* DisplayObj.SetCursor *)
  x:= max ( 0, min ( Size.x - 1, x ) );
  y:= max ( 0, min ( Size.y - 1, y ) );
  case WhichSystem of
    systemDOS, systemDPMI:
      begin
        dx:= ( ( y and $7F ) shl 8 ) or ( x and $7F );
        asm ( 'pushl %eax' );
        asm ( 'pushl %ebx' );
        asm ( 'pushl %edx' );
        asm ( 'movw %0, %%dx' : : 'rm' ( dx ) );
        asm ( 'movb $0x02, %ah' );
        asm ( 'xorb %bh, %bh' );
        asm ( 'int $0x10' );
        asm ( 'popl %edx' );
        asm ( 'popl %ebx' );
        asm ( 'popl %eax' );
      end (* systemDOS, systemDPMI *);
    systemOS2:
      begin
        VioSetCurType ( StartupCursorInfo, 0 );
        VioSetCurPos ( y, x, 0 );
      end (* systemOS2 *);
    otherwise:
      EMXinvalidOS;
  end (* case *);
  Cursor.x:= x;
  Cursor.y:= y;
end (* DisplayObj.SetCursor *);


Procedure DisplayObj.HideCursor;

Var
  dx: ShortWord;
  CursorInfo: VioCursorInfo;

begin (* DisplayObj.HideCursor *)
  (* Currently, the cursor is moved out of the visible screen. *)
  (* This will be changed as a "SetCursorStyle" procedure will *)
  (* be implemented -- then an invisible "cursor style" will   *)
  (* be used instead and this procedure will become a simple   *)
  (* call of "SetCursorStyle ( csNoCursor )".                  *)
  case WhichSystem of
    systemDOS, systemDPMI:
      begin
        dx:= 50 shl 8;
        asm ( 'pushl %eax' );
        asm ( 'pushl %ebx' );
        asm ( 'pushl %edx' );
        asm ( 'movw %0, %%dx' : : 'rm' ( dx ) );
        asm ( 'movb $0x02, %ah' );
        asm ( 'xorb %bh, %bh' );
        asm ( 'int $0x10' );
        asm ( 'popl %edx' );
        asm ( 'popl %ebx' );
        asm ( 'popl %eax' );
      end (* systemDOS, systemDPMI *);
    systemOS2:
      begin
        with CursorInfo do
          begin
            yStart:= 0;
            cEnd:= 0;
            cx:= 0;
            Attr:= $FFFF;
          end (* with *);
        VioSetCurType ( CursorInfo, 0 );
      end (* systemOS2 *);
    otherwise:
      EMXinvalidOS;
  end (* case *);
end (* DisplayObj.HideCursor *);


Procedure DisplayObj.WriteChars ( x, y, w, h: Integer;
                                  C: Char; Attr: Byte );

Var
  Buffer: array [ 0..1023 ] of Byte;

begin (* DisplayObj.WriteChars *)
  if x < 0 then
    begin
      inc ( w, x );
      x:= 0;
    end (* if *);
  w:= min ( Size.x - x, w );
  if y < 0 then
    begin
      inc ( h, y );
      y:= 0;
    end (* if *);
  h:= min ( Size.y - y, h );
  if ( w > 0 ) and ( h > 0 ) then
    begin
      FillShortWord ( Buffer, ( w + CharSize.x - 1 ) div CharSize.x,
                      ShortWord ( Attr ) shl 8 or ord ( C ) );
      inc ( h, y );
      while y < h do
        begin
          WriteBuffer ( x, y, w, Buffer );
          inc ( y, CharSize.y );
        end (* while *);
    end (* if *);
end (* DisplayObj.WriteChars *);


Procedure DisplayObj.WriteString ( x, y: Integer;
                                   S: WrkString; Attr: Byte );

Var
  Buffer: array [ 0..1023 ] of Byte;
  i, j: Integer;

begin (* DisplayObj.WriteString *)
  if ( length ( S ) > 0 ) then
    begin
      j:= 0;
      for i:= 1 to length ( S ) do
        begin
          Buffer [ j ]:= ord ( S [ i ] );
          inc ( j );
          Buffer [ j ]:= Attr;
          inc ( j );
        end (* for *);
      WriteBuffer ( x, y, length ( S ) * CharSize.x, Buffer );
    end (* if *);
end (* DisplayObj.WriteString *);


Procedure DisplayObj.WriteBuffer ( x, y, w: Integer; Var B: Void );

(* This procedure is awfully slow.  I will improve it *)
(* as soon as I have more experience with the GNU     *)
(* inline assembler.  (Borland's asm is much more     *)
(* comfortable!)                                      *)

Var
  (*$W-*)
  Buffer: array [ 0..511 ] of ShortWord absolute B;
  (*$W+*)
  D: ShortWord;
  O, addr, len: Word;
  P: ShortWordPtr;
  i0, i, xx: Integer;

begin (* DisplayObj.WriteBuffer *)
  if ( y < 0 ) or ( y >= Size.y ) then
    Exit;
  (* Do the clipping -- must be rewritten for graphics mode *)
  if x < 0 then
    begin
      i0:= - x;
      inc ( w, x );
      x:= 0;
    end (* if *)
  else
    i0:= 0;
  w:= min ( w, Size.x - x );
  case WhichSystem of
    systemDOS, systemDPMI:
      begin
        if SegVideo = 0 then
          FatalError ( 'WriteBuffer (EMX): cannot access video memory' );
        i:= i0;
        inc ( w, x );
        O:= 2 * ( Size.x * y + x );
        xx:= x;
        while xx < w do
          begin
            D:= Buffer [ i ];
            if WhichSystem = systemDOS then
              begin
                P:= Pointer ( SegVideo + O );
                P^:= D;
              end (* if *)
            else
              begin
                asm ( 'pushl %eax' );
                asm ( 'pushl %ebx' );
                asm ( 'movl %0, %%ebx' : : 'rm' ( O ) );
                asm ( 'movw %0, %%ax' : : 'rm' ( D ) );
                asm ( 'pushw %es' );
                asm ( 'movw %0, %%es' : : 'rm' ( SegVideo ) );
                asm ( 'movw %ax, %es:(%ebx)' );
                asm ( 'popw %es' );
                asm ( 'popl %ebx' );
                asm ( 'popl %eax' );
              end (* else *);
            inc ( i );
            inc ( O, 2 );
            inc ( xx, CharSize.x );
          end (* while *);
      end (* systemDOS, systemDPMI *);
    systemOS2:
      VioWrtCellStr ( Buffer [ i0 ], 2 * w, y, x, 0 );
    otherwise:
      EMXinvalidOS;
  end (* case *);
end (* DisplayObj.WriteBuffer *);


(*$endif *)  (* DISPLAYS *)
