(*$I Tools    *)   (* As long as GPC does not compile Units   *)
(*$I Objects  *)   (* automatically, :-( I include them here. *)
(*$I Events   *)
(*$I Displays *)


Program Menu;  (* A simple sample menu.  Don't invest too     *)
               (* much work in such things:  BO5 will provide *)
               (* a "Menus" Unit soon.                        *)


(*$I CompOpts.icl *)  (* Reasonable compiler options.     *)
                      (* Each module should include this. *)


uses
  Events,    (* Keyboard handling *)
  Displays;  (* Screen handling   *)


Var
  TheEnd: Boolean value false;
  MenuStr: array [ 1..3 ] of String ( 10 );


Procedure Init;

begin (* Init *)
  MenuStr [ 1 ]:= 'File';  (* Why does variable initialization *)
  MenuStr [ 2 ]:= 'Help';  (* not work with strings?! :-(      *)
  MenuStr [ 3 ]:= 'Exit';
end (* Init *);


Procedure DisplayMenu;

Var
  i: Integer;

(* HideCursor (Displays) hides the cursor, if the terminal allows  *)
(* it.  The WriteChars and WriteString procedures (Displays) don't *)
(* use or move the cursor.                                         *)
(*                                                                 *)
(* Coordinates: upper left corner is (0,0), lower right corner is  *)
(* ( ScreenWidth - 1, ScreenHeight - 1 ).                          *)

begin (* DisplayMenu *)
  HideCursor;
  (* Fill the screen with '#' *)
  WriteChars ( 0, 0, ScreenWidth, ScreenHeight, '#', darkgray );
  (* Menu box.  Not optimized for slow terminals. *)
  WriteChars ( 2 * CharWidth, 2 * CharHeight,
               10 * CharWidth, 3 * CharHeight, ' ',
               yellow + brown * background );
  for i:= 1 to 3 do
    begin
      WriteString ( 4 * CharWidth, ( 1 + i ) * CharHeight,
                    MenuStr [ i ],
                    yellow + brown * background );
      WriteString ( 4 * CharWidth, ( 1 + i ) * CharHeight,
                    MenuStr [ i ] [ 1 ],
                    white + brown * background );  (* first character *)
    end (* for *)
end (* DisplayMenu *);


Procedure FileManager;

(* Why is everything multiplied by CharWidth/CharHeight?     *)
(* Well, I hope to be able to implement graphics mode soon,  *)
(* and then this will become important.  In text mode, both  *)
(* values are set to one, so it makes no difference.         *)
(*                                                           *)
(* The ReadKey function waits for a character to be typed in *)
(* (without using the cursor) and returns it.  Here we use   *)
(* it like a procedure and ignore the result; a "dummy"      *)
(* variable is not needed.  Since CompOpts.icl contains an   *)
(* {$X+} compiler switch, this is accepted without warning.  *)
(* (And since CompOpts.icl also contains an {$N+} switch,    *)
(* these nested comments are accepted without an error.)     *)

begin (* FileManager *)
  HideCursor;
  ClearScreen ( blue );
  WriteString ( 15 * CharWidth, 5 * CharHeight,
                'File manager', white + blue * background );
  WriteString ( 15 * CharWidth, 6 * CharHeight,
                '============', white + blue * background );
  WriteString ( 15 * CharWidth, 10 * CharHeight,
                '<Press any key to return to menu>',
                lightgray + blue * background );
  ReadKey;
end (* FileManager *);


Procedure DisplayHelp;

begin (* DisplayHelp *)
  HideCursor;
  ClearScreen ( green );
  WriteString ( 15 * CharWidth, 5 * CharHeight,
                'Help', yellow + green * background );
  WriteString ( 15 * CharWidth, 6 * CharHeight,
                '====', yellow + green * background );
  WriteString ( 15 * CharWidth, 10 * CharHeight,
                '<Press any key to return to menu>',
                black + green * background );
  ReadKey;
end (* DisplayHelp *);


Procedure Fini;

(* SetCursor moves the hardware cursor, so the "normal" *)
(* write starts at that position.                       *)

begin (* Fini *)
  ClearScreen ( black );
  SetCursor ( 5 * CharWidth, ScreenHeight - 3 * CharHeight );
  (* For the last line it would be ScreenHeight - CharHeight. *)
  writeln ( 'That''s all folks ...' );
  writeln;
  (* Now the cursor is in the last line. *)
end (* Fini *);


(* The character returned by ReadKey is converted *)
(* to lower case using the LoCase table (Tools).  *)

begin
  Init;
  repeat
    DisplayMenu;
    case LoCase [ ReadKey ] of
      'f': FileManager;
      'h': DisplayHelp;
      'e': TheEnd:= true;
    end (* case *);
  until TheEnd;
  Fini;
end.
