(* Linux.icl -- included by SysDep.icl *)


(* Lo-level driver routines for Linux systems with VGA display *)


(******************************************************************************)
(*                                                                            *)
(*  This file is part of BO5, the user interface library for GNU Pascal.      *)
(*                                                                            *)
(*  BO5 will be distributed as free software (probably under the terms of     *)
(*  the GNU Library General Public License) as soon as it will have reached   *)
(*  some state of completion.  Up to that time, you are allowed to use BO5    *)
(*  for non-commercial purposes only.  You may distribute verbatim copies of  *)
(*  BO5 as a whole, but only without -- even a nominal -- fee.  You may not   *)
(*  distribute parts or changed versions of BO5.  I stress that there is      *)
(*  absolutely NO WARRANTY for BO5 -- any risk of using BO5 is with you!      *)
(*                                                                            *)
(*  Copyright (C) 1995  Peter Gerwinski, Duesseldorfer Str. 35                *)
(*                      (with TeX: D\"usseldorfer Str. 35)                    *)
(*                      45145 Essen, Germany                                  *)
(*                      e-mail: pege@mail.theo-phys.uni-essen.de              *)
(*                                                                            *)
(******************************************************************************)


(* Dear experienced lo-level programmer!       *)
(*                                             *)
(* Please don't cry when you read this source. *)
(* It was written rather hastily and will be   *)
(* optimized later on.                         *)
(*                                             *)
(* I have problems with the GNU inline assem-  *)
(* bler.  What is the correct way to use it?   *)
(* Any help welcome! -- PG                     *)


(*$ifdef SYSTOOLS *)  (* This goes into the Interface section of Tools *)


Var
  MemoryHandle: Integer;
  segA000: Word value 0;  (* Video memory *)
  segB000: Word value 0;
  segB800: Word value 0;


(*$endif *)  (* SYSTOOLS *)


(*$ifdef TOOLS *)  (* This goes into the Implementation section of Tools *)


(* Linux lo-level I/O control *)

Function ioCtl ( fildes, cmd: Integer; ... ): Integer; C;

Function ioPerm ( from, num: Word; turnon: Integer ): Integer; C;


(* Linux lo-level file system access (in fact: memory access) *)

Function mmap ( addr: Pointer; len: Word;
                prot, flags, fd, off: Integer ): Word; C;

(* In fact it returns a pointer.  But we convert it to Word. *)

Function open ( filename: CString; Flags: Integer; ... ): Integer; C;


Procedure InitTools;

Var
  Dummy: Word;


Function Allocate ( address, length: Word ): Word;

Var
  segMem: Pointer;
  result: Word;

begin (* Allocate *)
  segMem:= GetMem ( length + $FFF );
  if Word ( segMem ) and $FFF <> 0 then
    inc ( Word ( segMem ), $1000 - Word ( segMem ) and $FFF );
  if segMem = Nil then
    FatalError ( 'Linux drivers: cannot allocate video memory' );
  result:= mmap ( segMem, length, $03, $11, MemoryHandle, address );
  if Integer ( result ) <= 0 then
    FatalError ( 'Linux drivers: cannot mmap video memory' );
  Allocate:= result;
end (* Allocate *);


begin (* InitTools *)
  WhichSystem:= systemLinux;
  if ioCtl ( 0, $4B31, @Dummy ) = 0 then
    (* This is a console.  (This information will be stored, one day.) *);
  MemoryHandle:= open ( '/dev/mem', 2 );
  if MemoryHandle < 0 then
    FatalError ( 'Linux drivers: cannot open /dev/mem' );
  segA000:= Allocate ( $A0000, $10000 );
  segB000:= Allocate ( $B0000, $8000 );
  segB800:= Allocate ( $B8000, $8000 );
  if ioPerm ( $3D4, 1, 1 ) <> 0 then
    FatalError ( 'Linux drivers: cannot get I/O permissions' );
  ioPerm ( $3D5, 1, 1 );  (* must work now *)
end (* InitTools *);


Procedure FiniTools;

begin (* FiniTools *)
  (* empty *)
end (* FiniTools *);


(*$endif *)  (* TOOLS *)


(*$ifdef EVENTS *)


(* Linux keyboard access *)


(* The following declarations are collected from various C header files. *)
(* I know that it is superfluous work to redo everything in Pascal, but  *)
(* I want to demonstrate that it is possible without any problem to go   *)
(* into such deep regions of the machine with GNU Pascal.                *)
(*                                                                       *)
(* So let's hope that they won't change these constants and functions.   *)
(* If they prove to be too "volatile" then either this part of BO5 must  *)
(* be rewritten in C :-( or Linux must be rewritten in Pascal :-).       *)


Const
  IgnBrk  = $0001;
  BrkInt  = $0002;
  IgnPar  = $0004;
  ParMrk  = $0008;
  InPck   = $0010;
  IStrip  = $0020;
  InlCR   = $0040;
  IgnCR   = $0080;
  ICrnl   = $0100;
  IUClc   = $0200;
  IXon    = $0400;
  IXany   = $0800;
  IXoff   = $1000;
  IMaxBel = $2000;

  OPost   = $0001;

  ISig    = $0001;
  ICanon  = $0002;
  XCase   = $0004;
  Echo    = $0008;
  EchoE   = $0010;
  EchoK   = $0020;
  EchoNL  = $0040;
  NoFlsh  = $0080;
  ToStop  = $0100;
  EchoCtl = $0200;

  CRead   = $0080;

  TCSADRAIN = 1;


(* The ideas for the following lo-level keyboard routines for Linux *)
(* are stolen from the excellent C++ class library SPUNK by Ullrich *)
(* von Bassewitz, uz@ibb.schwaben.de, file kbd.cc.  The copyright   *)
(* notice in the file "hexed.doc" distributed with SPUNK allows me  *)
(* to do "anything what I want" with the SPUNK source except some   *)
(* things I would never have the idea to do. :-)  So this is legal. *)
(*                                                                  *)
(* (BO5 will access the keyboard in semi-raw mode as soon as I have *)
(* time to implement it.  So the similarity to SPUNK will decrease  *)
(* in future releases.)                                             *)


Type
  fdSet = set of 0..255;
  fdSetPtr = ^fdSet;


TimeVal = record
  sec, usec: Integer;
end (* TimeVal *);


TermIOS = record
  iFlag, oFlag, cFlag, lFlag: Word;
  line: Byte;
  cc: array [ 0..18 ] of Byte;
end (* TermIOS *);


Var
  StartupSettings: TermIOS;
  StartupKbdMode: Integer value -1;
  StartupMetaMode: Integer value -1;


Function ioCtl ( fildes, cmd: Integer; ... ): Integer; C;

Function tcGetAttr ( fildes: Integer; Var T: TermIOS ): Integer; C;

Function tcSetAttr ( fildes, OptionalActions: Integer;
                     Var T: TermIOS ): Integer; C;

Function select ( width: Integer; ReadFDS, WriteFDS, ExceptFDS: fdSetPtr;
                  Var TimeOut: TimeVal ): Integer; C;

Function CreadFD ( fd: Integer; Var Buffer: Void;
                   nbytes: Integer ): Integer; asmname 'read';


Function KeyPressed: Boolean;

Var
  TimeOut: TimeVal;
  Desc: fdSet;

begin (* KeyPressed *)
  with TimeOut do
    begin
      sec:= 0;
      usec:= 0;
    end (* with *);
  Desc:= [ 0 ];
  KeyPressed:= select ( 1, @Desc, Nil, Nil, TimeOut ) <> 0;
end (* KeyPressed *);


Function ReadKey: Char;

(* Read everything, but return only the first character.   *)
(* This yields simply ESC for F-keys, arrows, etc..        *)
(*                                                         *)
(* This will be rewritten completely, when BO5 will access *)
(* the keyboard in semi-raw mode.                          *)

Var
  Buffer: packed array [ 0..1023 ] of Char;

begin (* ReadKey *)
  repeat
  until KeyPressed;
  CreadFD ( 0, Buffer, SizeOf ( Buffer ) - 1 );
  ReadKey:= Buffer [ 0 ];
end (* ReadKey *);


Procedure InitEvents;

Var
  Dummy: Integer;
  Settings: TermIOS;

begin (* InitEvents *)
  tcGetAttr ( 0, StartupSettings );
  Settings:= StartupSettings;
  with Settings do
    begin
      and ( iFlag, not ( IgnBrk or IgnCR or InlCR or ICrnl or IUClc
                         or IXany or IXon or IXoff or InPck or IStrip ) );
      or ( iFlag, BrkInt or IgnPar );
      and ( lFlag, not ( XCase or EchoNL or NoFlsh or ICanon or ISig
                         or Echo ) );
      or ( cFlag, CRead );
      cc [ 5 ]:= 3;  (* Wait 300 ms for keys *)
      cc [ 6 ]:= 0;  (* return if timeout on keys available *)
    end (* with *);
  tcSetAttr ( 0, TCSADRAIN, Settings );
  if ioCtl ( 0, $4B62, @StartupMetaMode ) = 0 then
    ioCtl ( 0, $4B63, $04 )
  else
    StartupMetaMode:= - 1;
end (* InitEvents *);


Procedure FiniEvents;

begin (* FiniEvents *)
  tcSetAttr ( 0, TCSADRAIN, StartupSettings );
  if StartupMetaMode <> -1 then
    ioCtl ( 0, $4B63, StartupMetaMode );
end (* FiniEvents *);


(*$endif *)  (* EVENTS *)


(*$ifdef DISPLAYS *)


Function CwriteFD ( Handle: Integer; Var Buffer: Void; length: Word ): Integer;
AsmName 'write';

 
Var
  SegVideo: Word value 0;


Procedure WriteToTerminal ( S: WrkString );

begin (* WriteToTerminal *)
  if length ( S ) > 0 then
    CwriteFD ( 1, S [ 1 ], length ( S ) );
end (* WriteToTerminal *);


Constructor DisplayObj.Init;

begin (* DisplayObj.Init *)
  Size.x:= 80;
  Size.y:= 25;
  CharSize.x:= 1;
  CharSize.y:= 1;
  SegVideo:= segB800;
end (* DisplayObj.Init *);


Destructor DisplayObj.Fini;

begin (* DisplayObj.Fini *)
  (* empty *)
end (* DisplayObj.Fini *);


Procedure DisplayObj.Clear ( Attr: Byte );

(* Attr specifies a background color, so we select *)
(* a reasonable foreground color.                  *)
(* This only concerns subsequent write calls the   *)
(* application program should not use anyway.      *)
(* But if a fatal error happens, the message will  *)
(* be visible.                                     *)

Var
  ForeGround, BackGround: Byte;

begin (* DisplayObj.Clear *)
  BackGround:= Attr and $0F;
  ForeGround:= ( Attr and $F0 ) shr 4;
  if ( ForeGround = black ) and ( BackGround < lightgray ) then
    ForeGround:= lightgray;
  WriteChars ( 0, 0, Size.x, Size.y, ' ',
               ForeGround or BackGround shl 4 );
end (* DisplayObj.Clear *);


Procedure DisplayObj.SetCursor ( x, y: Integer );

Var
  xStr, yStr: String ( 10 );

(* Normally, WriteStr would convert Integers to Strings. *)
(* But you have to calculate the length of the String    *)
(* manually with GPC :-( and pass it after the ":", so   *)
(* it is not much more work to do everything manually.   *)

begin (* DisplayObj.SetCursor *)
  x:= max ( 1, min ( Size.x, x + 1 ) );
  y:= max ( 1, min ( Size.y, y + 1 ) );
  xStr:= '';
  while x > 0 do
    begin
      xStr:= chr ( ord ( '0' ) + x mod 10 ) + xStr;
      x:= x div 10;
    end (* while *);
  yStr:= '';
  while y > 0 do
    begin
      yStr:= chr ( ord ( '0' ) + y mod 10 ) + yStr;
      y:= y div 10;
    end (* while *);
  WriteToTerminal ( chr ( 27 ) + '[?25h' +
                    chr ( 27 ) + '[' + yStr + ';' + xStr + 'H' );
  Cursor.x:= x - 1;
  Cursor.y:= y - 1;
end (* DisplayObj.SetCursor *);


Procedure DisplayObj.HideCursor;

begin (* DisplayObj.HideCursor *)
  WriteToTerminal ( chr ( 27 ) + '[?25l' );
end (* DisplayObj.HideCursor *);


Procedure DisplayObj.WriteChars ( x, y, w, h: Integer;
                                  C: Char; Attr: Byte );

Var
  Buffer: array [ 0..1023 ] of Byte;

begin (* DisplayObj.WriteChars *)
  if x < 0 then
    begin
      inc ( w, x );
      x:= 0;
    end (* if *);
  w:= min ( Size.x - x, w );
  if y < 0 then
    begin
      inc ( h, y );
      y:= 0;
    end (* if *);
  h:= min ( Size.y - y, h );
  if ( w > 0 ) and ( h > 0 ) then
    begin
      FillShortWord ( Buffer, ( w + CharSize.x - 1 ) div CharSize.x,
                      ShortWord ( Attr ) shl 8 or ord ( C ) );
      inc ( h, y );
      while y < h do
        begin
          WriteBuffer ( x, y, w, Buffer );
          inc ( y, CharSize.y );
        end (* while *);
    end (* if *);
end (* DisplayObj.WriteChars *);


Procedure DisplayObj.WriteString ( x, y: Integer;
                                   S: WrkString; Attr: Byte );

Var
  Buffer: array [ 0..1023 ] of Byte;
  i, j: Integer;

begin (* DisplayObj.WriteString *)
  if ( length ( S ) > 0 ) then
    begin
      j:= 0;
      for i:= 1 to length ( S ) do
        begin
          Buffer [ j ]:= ord ( S [ i ] );
          inc ( j );
          Buffer [ j ]:= Attr;
          inc ( j );
        end (* for *);
      WriteBuffer ( x, y, length ( S ) * CharSize.x, Buffer );
    end (* if *);
end (* DisplayObj.WriteString *);


Procedure DisplayObj.WriteBuffer ( x, y, w: Integer; Var B: Void );

Var
  (*$W-*)
  Buffer: array [ 0..511 ] of ShortWord absolute B;
  (*$W+*)
  P: ShortWordPtr;
  Offset: ShortWord;
  i0, i, xx: Integer;

begin (* DisplayObj.WriteBuffer *)
  if ( y < 0 ) or ( y >= Size.y ) then
    Exit;
  (* Do the clipping -- must be rewritten for graphics mode *)
  if x < 0 then
    begin
      i0:= - x;
      inc ( w, x );
      x:= 0;
    end (* if *)
  else
    i0:= 0;
  w:= min ( w, Size.x - x );
  if SegVideo = 0 then
    FatalError ( 'WriteBuffer (Linux): cannot access video memory' );
  (* Linux consoles have a "dynamic" video base address. *)
  (* Read it out of the VGA registers now.               *)
  asm ( 'pushl %eax' );
  asm ( 'pushl %ebx' );
  asm ( 'pushl %edx' );
  asm ( 'movw $0x3D4, %dx' );  (* only correct in color mode *)
  asm ( 'movb $0x0C, %al' );
  asm ( 'outb %al, %dx' );
  asm ( 'incw %dx' );
  asm ( 'inb %dx, %al' );
  asm ( 'movb %al, %bh' );
  asm ( 'decw %dx' );
  asm ( 'movb $0x0D, %al' );
  asm ( 'outb %al, %dx' );
  asm ( 'incw %dx' );
  asm ( 'inb %dx, %al' );
  asm ( 'movb %al, %bl' );
  asm ( 'movw %%bx, %0' : : 'rm' ( Offset ) );
  asm ( 'popl %edx' );
  asm ( 'popl %ebx' );
  asm ( 'popl %eax' );
  i:= i0;
  inc ( w, x );
  P:= Pointer ( SegVideo + 2 * ( Size.x * y + x + Offset ) );
  xx:= x;
  while xx < w do
    begin
      P^:= Buffer [ i ];
      inc ( i );
      inc ( Word ( P ), 2 );
      inc ( xx, CharSize.x );
    end (* while *);
end (* DisplayObj.WriteBuffer *);


(*$endif *)  (* DISPLAYS *)
