Unit Objects;


(* Basic objects: collections, streams, ... *)


(******************************************************************************)
(*                                                                            *)
(*  This file is part of BO5, the user interface library for GNU Pascal.      *)
(*                                                                            *)
(*  BO5 will be distributed as free software (probably under the terms of     *)
(*  the GNU Library General Public License) as soon as it will have reached   *)
(*  some state of completion.  Up to that time, you are allowed to use BO5    *)
(*  for non-commercial purposes only.  You may distribute verbatim copies of  *)
(*  BO5 as a whole, but only without -- even a nominal -- fee.  You may not   *)
(*  distribute parts or changed versions of BO5.  I stress that there is      *)
(*  absolutely NO WARRANTY for BO5 -- any risk of using BO5 is with you!      *)
(*                                                                            *)
(*  Copyright (C) 1995  Peter Gerwinski, Duesseldorfer Str. 35                *)
(*                      (with TeX: D\"usseldorfer Str. 35)                    *)
(*                      45145 Essen, Germany                                  *)
(*                      e-mail: pege@mail.theo-phys.uni-essen.de              *)
(*                                                                            *)
(******************************************************************************)


(*$I CompOpts.icl *)


(* For the moment, please ignore this Unit. *)
(* The only implemented object is BaseObj   *)
(* which is nothing but a common parent of  *)
(* some (not yet existing) useful other     *)
(* objects.                                 *)


Interface


uses
  Tools;


Type
  BasePtr = ^BaseObj;

  (*$ifdef NOTYET *)
  CollectionPtr = ^CollectionObj;
  StreamPtr = ^StreamObj;
  (*$endif *)  (* NOTYET *)

  BasePtrArray = array [ 0..$FFFFFFF ] of BasePtr;
  BasePtrArrayPtr = ^BasePtrArray;


(* Abstract parent object for everything useful *)

BaseObj = object
  BeginData: record
    (* empty record just to mark the *)
    (* beginning of data fields      *)
  end (* BeginData *);
  Constructor Init;  (* Initialize everything to zero *)
  Destructor Fini; virtual;
end (* BaseObj *);


(*$ifdef NOTYET *)


(* Collection object: pointer to array of pointers to objects *)

CollectionObj = object ( BaseObj )
  Count: Integer;
  Data: BasePtrArrayPtr;
  Constructor Init;
  Destructor Fini; virtual;
end (* CollectionObj *);


(* Polymorphic I/O object *)

StreamObj = object ( BaseObj )
  Constructor Init;
  Destructor Fini; virtual;
  Procedure Read ( Var Buffer: Void; Size: Integer ); virtual;
  Procedure Write ( Var Buffer: Void; Size: Integer ); virtual;
  Function ReadObj: BasePtr;
  Procedure WriteObj ( P: BasePtr );
end (* StreamObj *);


(*$endif *)  (* NOTYET *)


Implementation


(*$I Base.met *)

(*$ifdef NOTYET *)
(*$I Collection.met *)
(*$I Stream.met     *)
(*$endif *)  (* NOTYET *)

end.
