(* TermCap.icl -- included by SysDep.icl *)


(* Lo-level driver routines for UNIX systems with TermCap display library *)


(******************************************************************************)
(*                                                                            *)
(*  This file is part of BO5, the user interface library for GNU Pascal.      *)
(*                                                                            *)
(*  BO5 will be distributed as free software (probably under the terms of     *)
(*  the GNU Library General Public License) as soon as it will have reached   *)
(*  some state of completion.  Up to that time, you are allowed to use BO5    *)
(*  for non-commercial purposes only.  You may distribute verbatim copies of  *)
(*  BO5 as a whole, but only without -- even a nominal -- fee.  You may not   *)
(*  distribute parts or changed versions of BO5.  I stress that there is      *)
(*  absolutely NO WARRANTY for BO5 -- any risk of using BO5 is with you!      *)
(*                                                                            *)
(*  Copyright (C) 1995  Peter Gerwinski, Duesseldorfer Str. 35                *)
(*                      (with TeX: D\"usseldorfer Str. 35)                    *)
(*                      45145 Essen, Germany                                  *)
(*                      e-mail: pege@mail.theo-phys.uni-essen.de              *)
(*                                                                            *)
(******************************************************************************)


(*$ifdef SYSTOOLS *)  (* This goes into the Interface section of Tools *)


(* empty *)


(*$endif *)  (* SYSTOOLS *)


(*$ifdef TOOLS *)  (* This goes into the Implementation section of Tools *)


Procedure InitTools;

begin (* InitTools *)
  WhichSystem:= systemTermCap;
end (* InitTools *);


Procedure FiniTools;

begin (* FiniTools *)
  (* empty *)
end (* FiniTools *);


(*$endif *)  (* TOOLS *)


(*$ifdef EVENTS *)


(* UNIX keyboard access *)


(* The following declarations are collected from various C header files. *)
(* I know that it is superfluous work to redo everything in Pascal, but  *)
(* I want to demonstrate that it is possible without any problem to go   *)
(* into such deep regions of the machine with GNU Pascal.                *)
(*                                                                       *)
(* So let's hope that they won't change these constants and functions.   *)
(* If they prove to be too "volatile" then either this part of BO5 must  *)
(* be rewritten in C :-( or Linux must be rewritten in Pascal :-).       *)


Const
  IgnBrk  = $0001;
  BrkInt  = $0002;
  IgnPar  = $0004;
  ParMrk  = $0008;
  InPck   = $0010;
  IStrip  = $0020;
  InlCR   = $0040;
  IgnCR   = $0080;
  ICrnl   = $0100;
  IUClc   = $0200;
  IXon    = $0400;
  IXany   = $0800;
  IXoff   = $1000;
  IMaxBel = $2000;

  OPost   = $0001;

  ISig    = $0001;
  ICanon  = $0002;
  XCase   = $0004;
  Echo    = $0008;
  EchoE   = $0010;
  EchoK   = $0020;
  EchoNL  = $0040;
  NoFlsh  = $0080;
  ToStop  = $0100;
  EchoCtl = $0200;

  CRead   = $0080;

  TIOCGWINSZ = $5413;

  TCSADRAIN = 1;

(* The ideas in this Unit are stolen from the excellent C++ class   *)
(* library SPUNK by Ullrich von Bassewitz, uz@ibb.schwaben.de.  The *)
(* copyright notice in the file "hexed.doc" distributed with SPUNK  *)
(* allows me to do "anything what I want" with the SPUNK source     *)
(* except some things I would never have the idea to do. :-)        *)
(* So this is legal.                                                *)
(*                                                                  *)
(* (I only *hope* that this will work on an arbitrary UNIX system,  *)
(* not only with Linux.  Since I cannot test it myself, please mail *)
(* me your test results.)                                           *)


Type
  fdSet = set of 0..255;
  fdSetPtr = ^fdSet;


TimeVal = record
  sec, usec: Integer;
end (* TimeVal *);


TermIOS = record
  iFlag, oFlag, cFlag, lFlag: Word;
  line: Byte;
  cc: array [ 0..18 ] of Byte;
end (* TermIOS *);


Var
  StartupSettings: TermIOS;
  StartupKbdMode: Integer value -1;
  StartupMetaMode: Integer value -1;


Function ioCtl ( fildes, cmd: Integer; ... ): Integer; C;

Function tcGetAttr ( fildes: Integer; Var T: TermIOS ): Integer; C;

Function tcSetAttr ( fildes, OptionalActions: Integer;
                     Var T: TermIOS ): Integer; C;

Function select ( width: Integer; ReadFDS, WriteFDS, ExceptFDS: fdSetPtr;
                  Var TimeOut: TimeVal ): Integer; C;

Function CreadFD ( fd: Integer; Var Buffer: Void;
                   nbytes: Integer ): Integer; 
AsmName 'read';


Function KeyPressed: Boolean;

Var
  TimeOut: TimeVal;
  Desc: fdSet;

begin (* KeyPressed *)
  with TimeOut do
    begin
      sec:= 0;
      usec:= 0;
    end (* with *);
  Desc:= [ 0 ];
  KeyPressed:= select ( 1, @Desc, Nil, Nil, TimeOut ) <> 0;
end (* KeyPressed *);


Function ReadKey: Char;

(* Read everything, but return only the first character.   *)
(* This yields simply ESC for F-keys, arrows, etc..        *)

Var
  Buffer: packed array [ 0..1023 ] of Char;

begin (* ReadKey *)
  repeat
  until KeyPressed;
  CreadFD ( 0, Buffer, SizeOf ( Buffer ) - 1 );
  ReadKey:= Buffer [ 0 ];
end (* ReadKey *);


Procedure InitEvents;

Var
  Dummy: Integer;
  Settings: TermIOS;

begin (* InitEvents *)
  tcGetAttr ( 0, StartupSettings );
  Settings:= StartupSettings;
  with Settings do
    begin
      and ( iFlag, not ( IgnBrk or IgnCR or InlCR or ICrnl or IUClc
                         or IXany or IXon or IXoff or InPck or IStrip ) );
      or ( iFlag, BrkInt or IgnPar );
      and ( lFlag, not ( XCase or EchoNL or NoFlsh or ICanon or ISig
                         or Echo ) );
      or ( cFlag, CRead );
      cc [ 5 ]:= 3;  (* Wait 300 ms for keys *)
      cc [ 6 ]:= 0;  (* return if timeout on keys available *)
    end (* with *);
  tcSetAttr ( 0, TCSADRAIN, Settings );
  if ioCtl ( 0, $4B62, @StartupMetaMode ) = 0 then
    ioCtl ( 0, $4B63, $04 )
  else
    StartupMetaMode:= - 1;
end (* InitEvents *);


Procedure FiniEvents;

begin (* FiniEvents *)
  tcSetAttr ( 0, TCSADRAIN, StartupSettings );
  if StartupMetaMode <> -1 then
    ioCtl ( 0, $4B63, StartupMetaMode );
end (* FiniEvents *);


(*$endif *)  (* EVENTS *)


(*$ifdef DISPLAYS *)


Type
  TermCapString = String ( 31 );


Var
  TerminalName: WrkString;
  TermCapBuffer: array [ 0..2047 ] of Char;
  InitTerminal, FiniTerminal, 
    InitKeyboard, FiniKeyboard, 
    CursorMotion, ClearTerminal,
    CursorOff, CursorOn, CursorFat,
    UnderlineOn, UnderlineOff, 
    StandOutOn, StandOutOff,
    ReverseOn, BlinkOn, BoldOn,
    ResetAttributes: TermCapString;
  CursorIsOff, HasColors: Boolean value false;
  LastAttr: ShortWord value $FFFF;
   

Function CwriteFD ( Handle: Integer; Var Buffer: Void; length: Word ): Integer;
AsmName 'write';
 

(* Functions in the TermCap library *)

Function tGetEnt ( Var Buffer: Void; Entry: CString ): Integer; C;

Function tGetStr ( Entry: CString; Var Buffer: CharPtr ): CharPtr; C;

Function tGetNum ( Entry: CString ): Integer; C;

Function tGoto ( CursorMotionStr: CString; x, y: Integer ): CharPtr; C;


Procedure WriteToTerminal ( S: WrkString );

begin (* WriteToTerminal *)
  if length ( S ) > 0 then
    CwriteFD ( 1, S [ 1 ], length ( S ) );
end (* WriteToTerminal *);


Procedure SetAttr ( Attr: Byte );

Var
  SeparatorNeeded: Boolean;

begin (* SetAttr *)
  if Attr <> LastAttr then
    begin
      if HasColors then
        begin
          SeparatorNeeded:= false;
          WriteToTerminal ( chr ( 27 ) + '[' );
          if Attr and $0F <> LastAttr and $FF0F then
            begin
              if Attr and $08 <> 0 then
                WriteToTerminal ( '01;' )  (* intensity *)
              else
                WriteToTerminal ( '21;' );
              (* This should be done via a table, but it causes *)
              (* too much trouble with GPC in the moment. :-(   *)
              case Attr and $07 of
                0: WriteToTerminal ( '30' );
                1: WriteToTerminal ( '34' );
                2: WriteToTerminal ( '32' );
                3: WriteToTerminal ( '36' );
                4: WriteToTerminal ( '31' );
                5: WriteToTerminal ( '35' );
                6: WriteToTerminal ( '33' );
                7: WriteToTerminal ( '37' );
              end (* case *);
              SeparatorNeeded:= true;
            end (* if *);
          if Attr and $F0 <> LastAttr and $FFF0 then
            begin
              if SeparatorNeeded then
                WriteToTerminal ( ';' );
              if Attr and $80 <> 0 then
                WriteToTerminal ( '05;' )  (* blink *)
              else
                WriteToTerminal ( '25;' );
              case ( Attr shr 4 ) and $07 of
                0: WriteToTerminal ( '40' );
                1: WriteToTerminal ( '44' );
                2: WriteToTerminal ( '42' );
                3: WriteToTerminal ( '46' );
                4: WriteToTerminal ( '41' );
                5: WriteToTerminal ( '45' );
                6: WriteToTerminal ( '43' );
                7: WriteToTerminal ( '47' );
              end (* case *);
            end (* if *);
          WriteToTerminal ( 'm' );
        end (* if *)
      else if length ( ResetAttributes ) > 0 then
        begin
          (* No colors, but ResetAttributes string exists. *)
          WriteToTerminal ( ResetAttributes );
          if Attr and $07 = $01 then
            WriteToTerminal ( UnderlineOn );
          if Attr and $08 <> 0 then
            WriteToTerminal ( BoldOn );
          if Attr and $80 <> 0 then
            WriteToTerminal ( BlinkOn );
          if Attr and $70 > $10 then
            if length ( ReverseOn ) > 0 then
              WriteToTerminal ( ReverseOn )
            else
              WriteToTerminal ( StandOutOn );
        end (* else if *)
      else
        begin
          (* No colors, no ResetAttributes string. *)
          if Attr and $07 <> LastAttr and $FF07 then
            if Attr and $07 = 1 then
              WriteToTerminal ( UnderlineOn )
            else
              WriteToTerminal ( UnderlineOff );
          if Attr and $70 <> LastAttr and $FF70 then
            if Attr and $70 <> 0 then
              WriteToTerminal ( StandOutOn )
            else
              WriteToTerminal ( StandOutOff );
        end (* else *);
      LastAttr:= Attr;
    end (* if *);
end (* SetAttr *);


Procedure ReadTermCapString ( Entry: WrkString; Var Cap: TermCapString );

Var
  Buffer: array [ 0..1023 ] of Char;
  C: CharPtr;

begin (* ReadTermCapString *)
  C:= @Buffer [ 0 ];
  C:= tGetStr ( Entry + chr ( 0 ), C );
  Cap:= '';
  if C <> Nil then
    while C^ <> chr ( 0 ) do
      begin
        Cap:= Cap + C^;
        inc ( Word ( C ) );
      end (* while *);
end (* ReadTermCapString *);


Constructor DisplayObj.Init;

Var
  S: WrkString;

WinSize: record
  Rows, Cols, xPixel, yPixel: ShortWord;
end (* WinSize *);

begin (* DisplayObj.Init *)
  TerminalName:= GetEnv ( 'TERM' );
  if TerminalName = '' then
    FatalError ( 'TermCap drivers: TERM environment variable not set' );
  case tGetEnt ( TermCapBuffer, TerminalName + chr ( 0 ) ) of
    -1: FatalError ( 'TermCap drivers: terminal database "/etc/termcap" not found' );
    0:  FatalError ( 'TermCap drivers: terminal "' + TerminalName + 
                     '" not found in database' );
  end (* case *);
  ReadTermCapString ( 'cm', CursorMotion );
  if CursorMotion = '' then
    FatalError ( 'TermCap drivers: terminal has no random cursor access' );
  ReadTermCapString ( 'is', InitTerminal );
  ReadTermCapString ( 'rs', FiniTerminal ); 
  ReadTermCapString ( 'ks', InitKeyboard );
  ReadTermCapString ( 'ke', FiniKeyboard ); 
  ReadTermCapString ( 'cl', ClearTerminal );
  ReadTermCapString ( 'vi', CursorOff );
  ReadTermCapString ( 've', CursorOn );
  ReadTermCapString ( 'vs', CursorFat );
  ReadTermCapString ( 'us', UnderlineOn );
  ReadTermCapString ( 'ue', UnderlineOff ); 
  ReadTermCapString ( 'so', StandOutOn );
  ReadTermCapString ( 'se', StandOutOff );
  ReadTermCapString ( 'mr', ReverseOn );
  ReadTermCapString ( 'mb', BlinkOn );
  ReadTermCapString ( 'md', BoldOn );
  ReadTermCapString ( 'me', ResetAttributes );
  WriteToTerminal ( InitTerminal );
  WriteToTerminal ( InitKeyboard );
  (* Use the same strategy as SPUNK to determine display size. *)
  ioCtl ( 0, TIOCGWINSZ, @WinSize );
  with WinSize do
    begin
      Size.x:= Cols;
      Size.y:= Rows;
    end (* with *);
  (* If ioCtl was not successful, use the environment *)
  if ( Size.x = 0 ) or ( Size.y = 0 ) then
    begin
      S:= GetEnv ( 'COLUMNS' );
      if length ( S ) > 0 then
        ReadStr ( S, Size.x );
      S:= GetEnv ( 'LINES' );
      if length ( S ) > 0 then
        ReadStr ( S, Size.y );
    end (* if *);
  (* If this also failed, use TermCap *)
  if ( Size.x = 0 ) or ( Size.y = 0 ) then
    begin
      Size.x:= tgetnum ( 'co' );
      Size.y:= tgetnum ( 'li' );
    end (* if *);
  if ( Size.x = 0 ) or ( Size.y = 0 ) then
    FatalError ( 'TermCap drivers: cannot determine display size' );
  CharSize.x:= 1;
  CharSize.y:= 1;
  (* Do we have colors?  Check the CL_COLOR environment variable. *)
  S:= GetEnv ( 'CL_COLOR' );
  if ( S = '1' ) or ( S = 'ON' ) or ( S = 'YES' ) then
    HasColors:= true;
end (* DisplayObj.Init *);


Destructor DisplayObj.Fini;

begin (* DisplayObj.Fini *)
  SetAttr ( lightgray );
  WriteToTerminal ( FiniTerminal );
  WriteToTerminal ( FiniKeyboard );
end (* DisplayObj.Fini *);


Procedure DisplayObj.Clear ( Attr: Byte );

(* Attr specifies a background color, so we select *)
(* a reasonable foreground color.                  *)
(* This only concerns subsequent write calls the   *)
(* application program should not use anyway.      *)
(* But if a fatal error happens, the message will  *)
(* be visible.                                     *)

Var
  ForeGround, BackGround: Byte;

begin (* DisplayObj.Clear *)
  BackGround:= Attr and $0F;
  ForeGround:= ( Attr and $F0 ) shr 4;
  if ( ForeGround = black ) and ( BackGround < lightgray ) then
    ForeGround:= lightgray;
  if ( BackGround = black ) and ( length ( ClearTerminal ) > 0 ) then
    begin
      SetAttr ( ForeGround or BackGround shl 4 );
      WriteToTerminal ( ClearTerminal );
    end (* if *)
  else
    WriteChars ( 0, 0, Size.x, Size.y, ' ',
                 ForeGround or BackGround shl 4 );
end (* DisplayObj.Clear *);


Procedure DisplayObj.SetCursor ( x, y: Integer );

Var
  C: CharPtr;
  S: WrkString;

begin (* DisplayObj.SetCursor *)
  x:= max ( 0, min ( Size.x - 1, x ) );
  y:= max ( 0, min ( Size.y - 1, y ) );
  if CursorIsOff then
    begin
      WriteToTerminal ( CursorOn );
      CursorIsOff:= false;
    end (* if *);
  if ( x <> Cursor.x ) or ( y <> Cursor.y ) then
    begin
      C:= tGoto ( CursorMotion + chr ( 0 ), x, y );
      S:= '';
      if C <> Nil then
        while C^ <> chr ( 0 ) do
          begin
            S:= S + C^;
            inc ( Word ( C ) );
          end (* while *);
      WriteToTerminal ( S );
      Cursor.x:= x;
      Cursor.y:= y;
    end (* if *);
end (* DisplayObj.SetCursor *);


Procedure DisplayObj.HideCursor;

begin (* DisplayObj.HideCursor *)
  if not CursorIsOff then
    begin
      WriteToTerminal ( CursorOff );
      CursorIsOff:= true;
    end (* if *);
end (* DisplayObj.HideCursor *);


Procedure DisplayObj.WriteChars ( x, y, w, h: Integer;
                                  C: Char; Attr: Byte );

Var
  S: WrkString;
  i: Integer;
  SaveCursor: Point;
  CursorWasOff: Boolean;

begin (* DisplayObj.WriteChars *)
  if x < 0 then
    begin
      inc ( w, x );
      x:= 0;
    end (* if *);
  w:= min ( Size.x - x, w );
  if y < 0 then
    begin
      inc ( h, y );
      y:= 0;
    end (* if *);
  h:= min ( Size.y - y, h );
  if ( w > 0 ) and ( h > 0 ) then
    begin
      inc ( h, y );
      SaveCursor:= Cursor;
      CursorWasOff:= CursorIsOff;
      S:= '';
      for i:= 1 to w do
        S:= S + C;
      SetAttr ( Attr );
      while y < h do
        begin
          SetCursor ( x, y );
          HideCursor;
          WriteToTerminal ( S );
          inc ( y, CharSize.y );
        end (* while *);
      SetCursor ( SaveCursor.x, SaveCursor.y );
      if CursorWasOff then
        HideCursor;
    end (* if *);
end (* DisplayObj.WriteChars *);


Procedure DisplayObj.WriteString ( x, y: Integer;
                                   S: WrkString; Attr: Byte );

Var
  SaveCursor: Point;
  CursorWasOff: Boolean;

begin (* DisplayObj.WriteString *)
  if ( length ( S ) > 0 ) then
    begin
      SaveCursor:= Cursor;
      CursorWasOff:= CursorIsOff;
      SetAttr ( Attr );
      SetCursor ( x, y );
      HideCursor;
      WriteToTerminal ( S );
      SetCursor ( SaveCursor.x, SaveCursor.y );
      if CursorWasOff then
        HideCursor;
    end (* if *);
end (* DisplayObj.WriteString *);


Procedure DisplayObj.WriteBuffer ( x, y, w: Integer; Var B: Void );

Var
  (*$W-*)
  Buffer: array [ 0..1023 ] of Char absolute B;
  (*$W+*)
  i0, i, xx: Integer;
  SaveCursor: Point;
  CursorWasOff: Boolean;

begin (* DisplayObj.WriteBuffer *)
  if ( y < 0 ) or ( y >= Size.y ) then
    Exit;
  (* Do the clipping -- must be rewritten for graphics mode *)
  if x < 0 then
    begin
      i0:= - x;
      inc ( w, x );
      x:= 0;
    end (* if *)
  else
    i0:= 0;
  w:= min ( w, Size.x - x );
  SaveCursor:= Cursor;
  CursorWasOff:= CursorIsOff;
  SetCursor ( x, y );
  HideCursor;
  i:= i0;
  xx:= x;
  while xx < x + w do
    begin
      SetAttr ( ord ( Buffer [ 2 * i + 1 ] ) );
      WriteToTerminal ( Buffer [ 2 * i ] );
      inc ( i );
      inc ( xx, CharSize.x );
    end (* while *);
  SetCursor ( SaveCursor.x, SaveCursor.y );
  if CursorWasOff then
    HideCursor;
end (* DisplayObj.WriteBuffer *);


(*$endif *)  (* DISPLAYS *)
